BROKER_BIN_DIR="/usr/share/dcv-session-manager-broker/bin"

BROKER_DEFAULT_CONF_FILENAME="default-session-manager-broker.properties"
CA_PEM_NAME="dcvsmbroker_ca.pem"
CA_KEY_NAME="dcvsmbroker_ca.key"
CA_SRL_NAME="dcvsmbroker_ca.srl"
CA_CERT_SUBJECT="/C=US/CN=DCV Session Manager Broker CA"
SIGNER_SUFFIX=".csr"
PEM_SUFFIX=".pem"
KEY_SUFFIX=".key"
CERT_SUBJECT="/C=US/CN=$(hostname)"
CERT_DURATION_DAYS=730
KIND_B2B="broker2broker"
KIND_A2B="agent2broker"
KIND_C2B="client2broker"
KIND_G2B="gateway2broker"
JKS_PEM_ALIAS="dcvsm_autogenerated"

source "${BROKER_BIN_DIR}/common.sh"

function check_program {
    if "$@" >/dev/null 2>&1; then
        echo "valid"
    else
        echo "Program '$*' is missing" >&2
    fi
}

function is_cert_valid {
    local _pem="$1"
    local _now
    local _format_opt
    local _expiration=0
    local _days_before_expiry="40"
    local _seconds_before_expiry=$(( _days_before_expiry * 24 * 60 * 60 ))

    _now=$(date +%s)
    [ "$2" ] && _format_opt=("-inform" "$2")

    IFS="=" read -ra _end_date <<< "$(openssl x509 -noout -enddate -in "${_pem}" "${_format_opt[@]}")"
    [ "${_end_date[1]}" ] && _expiration=$(date -d "${_end_date[1]}" +%s)

    echo "Checking cert validity for ${_pem}" >&2
    echo "Expiry time: ${_expiration}" >&2
    echo "Current time: ${_now}" >&2
    echo "Seconds before expiry: $(( _expiration - _now ))" >&2
    echo "Max seconds before expiry: ${_seconds_before_expiry}" >&2
    # Check if certificate expires within specified days
    if [ $(( _expiration - _now )) -gt $_seconds_before_expiry ]; then
        echo "Certificate is valid" >&2
        echo "valid"
    else
        if [ $(( _expiration - _now )) -le $_seconds_before_expiry ]; then
          local _days_until_expiry=$(( (_expiration - _now) / 86400 ))
          echo "Certificate will expire in $_days_until_expiry days" >&2
          echo "Restart broker to apply new certificate: sudo systemctl restart dcv-session-manager-broker" >&2
        fi
    fi
}

function stored_cert_hostname_matches {
    local _pem="$1"
    local _hostname

    _hostname="$(hostname)"
    openssl x509 -in "${_pem}" -inform DER -checkhost "${_hostname}" -noout | grep "does match"

}

function is_cert_issued_by_ca_cert {
    local _der="$1"
    local _ca_pem="$2"
    local _pem="${_der}.fromder"

    if openssl x509 -inform der -in "${_der}" -out "${_pem}" >/dev/null 2>&1; then
        if openssl verify -CAfile "${_ca_pem}" "${_pem}" >/dev/null 2>&1; then
            echo "valid"
        fi
        rm -f "${_pem}" 2>/dev/null 2>&1
    fi
}

function check_ca_cert {
    local _ca_key="$1"
    local _ca_pem="$2"

    if [ -r "${_ca_pem}" ] && [ -r "${_ca_key}" ] && [ "$(is_cert_valid "${_ca_pem}")" ]; then
        echo "valid"
    fi
}

function is_ipv4 {
    local _ip="$1"

    [[ "${_ip}" =~ ^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$ ]] && echo "${_ip}"
}

function get_ip_from_hostname {
    local _ips
    local _ip

    _ip="$(host -4 "$(hostname)" | cut -d " " -f 4)"
    if [ "$(is_ipv4 "${_ip}")" ]; then
        echo "${_ip}"
    else
        read -ra _ips <<< "$(hostname -I)"
        _ip="${_ips[0]}"
        [ "$(is_ipv4 "${_ip}")" ] && echo "${_ip}"
    fi
}

function get_cert_config {
    local _constraint
    local _ip
    local _ip_alt_name

    _ip="$(get_ip_from_hostname)"
    if [ "${_ip}" ]; then
        _ip_alt_name=",IP:${_ip}"
    else
        echo "Cannot discover IPv4 address for host $(hostname), it will not be added in the certificate" >&2
    fi

    if [ "$1" = "ca" ]; then
        _constraint="
basicConstraints=critical,CA:TRUE,pathlen:0
keyUsage=critical,keyCertSign,cRLSign"
    else
        _constraint="
authorityKeyIdentifier=keyid,issuer
basicConstraints=CA:FALSE
keyUsage=digitalSignature,nonRepudiation,keyEncipherment,dataEncipherment
subjectAltName=DNS:$(hostname)${_ip_alt_name}"
    fi

    echo "
[req]
distinguished_name=dn
[dn]
[ext]
${_constraint}"
}

function create_ca_cert {
    local _key="$1"
    local _pem="$2"

    if openssl req \
            -new -x509 -nodes -newkey rsa:4096 -sha256 \
            -days ${CERT_DURATION_DAYS} \
            -subj "${CA_CERT_SUBJECT}" \
            -extensions ext \
            -config <(get_cert_config ca) \
            -keyout "${_key}" \
            -out "${_pem}"; then
        chmod 600 "${_key}"
        echo "CA certificate ${_pem} created successfully"
    fi
}

function create_cert {
    local _ca_key="$1"
    local _ca_pem="$2"
    local _ca_srl="$3"
    local _key="$4"
    local _pem="$5"
    local _signer="${_pem}${SIGNER_SUFFIX}"

    if openssl req \
            -new -nodes -newkey rsa:4096 -sha256 \
            -subj "${CERT_SUBJECT}" \
            -keyout "${_key}" \
            -out "${_signer}"; then
        if openssl x509 \
                -req \
                -in "${_signer}" \
                -CA "${_ca_pem}" \
                -CAkey "${_ca_key}" \
                -CAcreateserial -sha256 -days ${CERT_DURATION_DAYS} \
                -extensions ext \
                -extfile <(get_cert_config pem) \
                -out "${_pem}"; then
             echo "Certificate ${_pem} created successfully"
             rm -f "${_ca_srl}" >/dev/null 2>&1
        fi
        rm -f "${_signer}" >/dev/null 2>&1
    fi
}

function generate_jks_password {
    < /dev/urandom tr -dc _A-Z-a-z-0-9 | head -c32
}

function save_jks_password {
    local _password_param="$1"
    local _password="$2"
    local _conf_file="$3"

    if < "${_conf_file}" grep -q "^[[:space:]]*${_password_param}.*"; then
        sed -i "s/^[[:space:]]*${_password_param}.*/${_password_param} = ${_password}/g" "${_conf_file}"
    else
        echo "${_password_param} = ${_password}" >> "${_conf_file}"
    fi
}

function add_certs_to_jks {
    local _ca_pem="$1"
    local _pem="$2"
    local _key="$3"
    local _jks_file="$4"
    local _jks_password="$5"
    local _p12="${_pem}.p12"

    if openssl pkcs12 \
            -export -chain \
            -in "${_pem}" \
            -inkey "${_key}" \
            -CAfile "${_ca_pem}" \
            -out "${_p12}" \
            -password pass:"${_jks_password}" \
            -name "${JKS_PEM_ALIAS}" -caname ca; then
        if keytool -importkeystore \
                -noprompt \
                -destkeystore "${_jks_file}" \
                -deststorepass "${_jks_password}" \
                -srckeystore "${_p12}" \
                -srcstoretype PKCS12 \
                -srcstorepass "${_jks_password}" \
                -alias "${JKS_PEM_ALIAS}"; then
            echo "Keystore ${_jks_file} updated successfully"
            chmod 600 "${_jks_file}"
        fi
        rm -f "${_pem}" "${_key}" >/dev/null 2>&1
    fi
    
    rm -f "${_p12}" >/dev/null 2>&1
}

function get_cert_from_jks {
    local _jks_file="$1"
    local _jks_password="$2"
    local _security_dir="$3"
    local _now
    local _cert

    _now=$(date +%s)
    _cert="${_security_dir}/jks_${_now}.pem"

    if keytool -exportcert \
            -alias "${JKS_PEM_ALIAS}" \
            -file "${_cert}" \
            -storepass "${_jks_password}" \
            -keystore "${_jks_file}" >/dev/null 2>&1; then
        echo "${_cert}"
    else
        rm -f "${_cert}" >/dev/null 2>&1
    fi
}

function clean_jks {
    local _jks_file="$1"

    rm -f "${_jks_file}" >/dev/null 2>&1
}

function create_cert_chain {
    local _ca_key="$1"
    local _ca_pem="$2"
    local _ca_srl="$3"
    local _key="$4"
    local _pem="$5"

    if [ "$(check_ca_cert "${_ca_key}" "${_ca_pem}")" ]; then
        echo "Autogenerated CA certificate ${_ca_pem} found alongside its private key, using them to create the certificate"
    else
        echo "Autogenerated CA certificate and private key not found or expired, creating new ones"
        create_ca_cert "${_ca_key}" "${_ca_pem}"
    fi

    create_cert "${_ca_key}" "${_ca_pem}" "${_ca_srl}" "${_key}" "${_pem}"
}

#
# Main function
#
function manage_certificate_kind {
    local _kind="$1"
    local _chroot_dir="$2"
    local _broker_conf_file="${_chroot_dir}/${BROKER_DEFAULT_CONF_FILENAME}"
    local _security_dir="${_chroot_dir}/security"
    local _ca_key_filename="${_security_dir}/${CA_KEY_NAME}"
    local _ca_pem_filename="${_security_dir}/${CA_PEM_NAME}"
    local _ca_srl_filename="${_security_dir}/${CA_SRL_NAME}"
    local _key_filename="${_security_dir}/dcvsmbroker_${_kind}${KEY_SUFFIX}"
    local _pem_filename="${_security_dir}/dcvsmbroker_${_kind}${PEM_SUFFIX}"
    local _jks_file_param
    local _jks_file
    local _jks_password
    local _jks_password_param
    local _stored_cert

    case "${_kind}" in
    "${KIND_B2B}")
        _jks_file_param="broker-to-broker-key-store-file"
        _jks_password_param="broker-to-broker-key-store-pass"
        ;;
    "${KIND_A2B}")
        _jks_file_param="agent-to-broker-connector-key-store-file"
        _jks_password_param="agent-to-broker-connector-key-store-pass"
        ;;
    "${KIND_C2B}")
        _jks_file_param="client-to-broker-connector-key-store-file"
        _jks_password_param="client-to-broker-connector-key-store-pass"
        ;;
    "${KIND_G2B}")
        _jks_file_param="gateway-to-broker-connector-key-store-file"
        _jks_password_param="gateway-to-broker-connector-key-store-pass"
        ;;
    *)
        echo "Invalid certificate kind"
        return
        ;;
    esac

    _jks_file=$(get_value_from_conf "${_broker_conf_file}" "${_jks_file_param}")
    if [ ! "${_jks_file}" ]; then
        _jks_file="${_security_dir}/keystore_${_kind}.jks"
        echo "${_jks_file_param} = ${_jks_file}" >> "${_broker_conf_file}"
    fi

    if [ ! -r "${_jks_file}" ]; then
        echo "No keystore ${_jks_file} found, creating a new one alongside needed certificates"
        create_cert_chain "${_ca_key_filename}" "${_ca_pem_filename}" "${_ca_srl_filename}" "${_key_filename}" "${_pem_filename}"
        _jks_password=$(generate_jks_password)
        add_certs_to_jks "${_ca_pem_filename}" "${_pem_filename}" "${_key_filename}" "${_jks_file}" "${_jks_password}"
        save_jks_password "${_jks_password_param}" "${_jks_password}" "${_broker_conf_file}"
    else
        echo "Found keystore ${_jks_file}"
        _jks_password=$(get_value_from_conf "${_broker_conf_file}" "${_jks_password_param}")
        _stored_cert=$(get_cert_from_jks "${_jks_file}" "${_jks_password}" "${_security_dir}")

        if [ ! -r "${_stored_cert}" ]; then
            echo "Unable to retrieve autogenerated certificate from keystore, nothing can be done"
            return
        fi

        if [ "$(is_cert_valid "${_stored_cert}" DER)" ] && [ "$(stored_cert_hostname_matches "${_stored_cert}")" ]; then
            echo "Not expired autogenerated certificate with matching hostname found in keystore ${_jks_file}"
            if [ "$(check_ca_cert "${_ca_key_filename}" "${_ca_pem_filename}")" ]; then
                echo "Valid autogenerated CA certificate ${_ca_pem_filename} found alongside its private key"
                echo "Checking that autogenerated certificate in keystore is issued by the autogenerated CA certificate..."
                if [ "$(is_cert_issued_by_ca_cert "${_stored_cert}" "${_ca_pem_filename}" )" ]; then
                    echo "Autogenerated stored certificate is issued by ${_ca_pem_filename}, all good"
                else
                    echo "Autogenerated stored certificate is NOT issued by ${_ca_pem_filename}, creating a new one"
                    create_cert "${_ca_key_filename}" "${_ca_pem_filename}" "${_ca_srl_filename}" "${_key_filename}" "${_pem_filename}"
                    clean_jks "${_jks_file}" "${_jks_password}"
                    add_certs_to_jks "${_ca_pem_filename}" "${_pem_filename}" "${_key_filename}" "${_jks_file}" "${_jks_password}"
                fi
            else
                echo "Autogenerated stored certificate is valid"
            fi
        else
            echo "Autogenerated stored certificate is expired, expiring soon or hostname does not match, creating new one"
            create_cert_chain "${_ca_key_filename}" "${_ca_pem_filename}" "${_ca_srl_filename}" "${_key_filename}" "${_pem_filename}"
            clean_jks "${_jks_file}" "${_jks_password}"
            add_certs_to_jks "${_ca_pem_filename}" "${_pem_filename}" "${_key_filename}" "${_jks_file}" "${_jks_password}"
        fi

        rm -f "${_stored_cert}" >/dev/null 2>&1
    fi
}

#
# Entry point
#
function manage_certificates {
    [ "$(check_program openssl version)" ] || return
    [ "$(check_program keytool)" ] || return

    local _broker_default_conf_file="${CHROOT_DIR}/${BROKER_DEFAULT_CONF_FILENAME}"
    local _params=("broker-to-broker-key-store-file" \
            "broker-to-broker-key-store-pass" \
            "agent-to-broker-connector-key-store-file" \
            "agent-to-broker-connector-key-store-pass" \
            "client-to-broker-connector-key-store-file" \
            "client-to-broker-connector-key-store-pass" \
            "gateway-to-broker-connector-key-store-file" \
            "gateway-to-broker-connector-key-store-pass")
    local _value

    for _param in "${_params[@]}"; do
        _value=$(get_value_from_conf "${BROKER_CONF_FILE}" "${_param}")
        if [ "${_value}" ]; then
            echo "Parameter ${_param} is set in ${BROKER_CONF_FILE}, skipping automatic certificate management"
            return
        fi
    done

    if [ ! -f "${_broker_default_conf_file}" ]; then
        touch "${_broker_default_conf_file}"
        chmod 640 "${_broker_default_conf_file}"
    fi

    for _kind in "${KIND_B2B}" "${KIND_A2B}" "${KIND_C2B}" "${KIND_G2B}"; do
        manage_certificate_kind "${_kind}" "${CHROOT_DIR}"
    done
}
